import logging
from typing import Optional

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration


def before_send(event: dict, hint: dict) -> Optional[dict]:
    if "exc_info" in hint:
        _, _, _ = hint["exc_info"]

    return event


def init_sentry(*, sentry_log_level: Optional[int], dsn: str) -> None:
    sentry_logging = LoggingIntegration(
        level=sentry_log_level,  # Capture info and above as breadcrumbs
        event_level=logging.ERROR,  # Send errors as events
    )
    integrations = [sentry_logging]

    sentry_sdk.init(
        dsn=dsn,
        integrations=integrations,
        environment="production",
        traces_sample_rate=0.0,
        send_default_pii=True,
        before_send=before_send,  # type: ignore
        max_request_body_size="always",
        max_value_length=10_000,
    )



import logging
import re
from unittest.mock import patch
from uuid import UUID

from workflow_server.server import create_app


def test_version_route():
    flask_app = create_app()

    with flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")
        assert response.status_code == 200
        assert re.match(r"[0-9]*\.[0-9]*\.[0-9]*", response.json["sdk_version"])
        assert response.json["server_version"] == "local"


def test_version_route__with_single_node_file(tmp_path):
    # GIVEN a temporary custom_nodes directory with a test node
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    node_file = custom_nodes_dir / "test_node.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class TestNode(BaseNode):
    \"""A test node for processing data.

    This is a detailed description of what the node does.
    \"""
    label = "Test Node"
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find exactly one node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the node should have the correct metadata
    node = nodes[0]
    assert UUID(node["id"])
    assert node["module"] == "vellum_custom_nodes"
    assert node["name"] == "TestNode"
    assert node["label"] == "Test Node"
    assert "A test node for processing data." in node["description"]
    assert "This is a detailed description" in node["description"]


def test_version_route__with_nodes_in_multiple_files(tmp_path):
    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND a first node file
    first_node_file = custom_nodes_dir / "first_node.py"
    first_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeNode(BaseNode):
    \"""This is Some Node.\"""
    label = "Some node"
"""
    )

    # AND a second node file
    second_node_file = custom_nodes_dir / "second_node.py"
    second_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeOtherNode(BaseNode):
    \"""This is Some Other Node.\"""
    label = "Some other node"
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find both nodes
    nodes = response.json["nodes"]
    assert len(nodes) == 2

    # AND the first node should have correct metadata
    some_node = nodes[0]
    assert some_node["label"] == "Some node"
    assert some_node["description"] == "This is Some Node."
    assert UUID(some_node["id"])
    assert some_node["module"] == "vellum_custom_nodes"

    # AND the second node should have correct metadata
    some_other_node = nodes[1]
    assert some_other_node["label"] == "Some other node"
    assert some_other_node["description"] == "This is Some Other Node."
    assert UUID(some_other_node["id"])
    assert some_other_node["module"] == "vellum_custom_nodes"


def test_version_route__no_custom_nodes_dir(tmp_path):
    # GIVEN a Flask application and an empty temp directory
    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND the nodes list should be empty
    assert response.json["nodes"] == []


def test_version_route__with_multiple_nodes_in_file(tmp_path):
    # Create a temporary custom_nodes directory with multiple nodes in one file
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # Create a test node file with multiple nodes
    node_file = custom_nodes_dir / "multiple_nodes.py"
    node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class ProcessingNode(BaseNode):
    \"""Processes input data.\"""
    label = "Processing Node"

class TransformationNode(BaseNode):
    \"""Transforms data format.\"""
    label = "Transformation Node"

# This class should not be discovered
class HelperClass:
    pass
"""
    )

    flask_app = create_app()

    # Mock the current working directory to point to our temp directory
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

        assert response.status_code == 200
        nodes = response.json["nodes"]
        assert len(nodes) == 2

        # Nodes should be discovered regardless of their order in the file
        node_names = {node["name"] for node in nodes}
        assert node_names == {"ProcessingNode", "TransformationNode"}


def test_version_route__with_invalid_node_file(tmp_path, caplog):
    caplog.set_level(logging.WARNING)

    # GIVEN a temporary custom_nodes directory
    custom_nodes_dir = tmp_path / "vellum_custom_nodes"
    custom_nodes_dir.mkdir()

    # AND a valid node file
    valid_node_file = custom_nodes_dir / "valid_node.py"
    valid_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class SomeNode(BaseNode):
    \"\"\"This is Some Node.\"\"\"
    label = "Some node"
"""
    )

    # AND an invalid node file with syntax error of missing colon in the class
    invalid_node_file = custom_nodes_dir / "invalid_node.py"
    invalid_node_file.write_text(
        """
from vellum.workflows.nodes import BaseNode

class BrokenNode(BaseNode)
    \"\"\"This node has a syntax error.\"\"\"
    label = "Broken Node"
"""
    )

    flask_app = create_app()

    # WHEN we make a request to the version route
    with patch("os.getcwd", return_value=str(tmp_path)), flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")

    # THEN we should get a successful response
    assert response.status_code == 200

    # AND we should find only the valid node
    nodes = response.json["nodes"]
    assert len(nodes) == 1

    # AND the valid node should have correct metadata
    valid_node = nodes[0]
    assert valid_node["label"] == "Some node"
    assert valid_node["description"] == "This is Some Node."
    assert UUID(valid_node["id"])
    assert valid_node["module"] == "vellum_custom_nodes"

    # AND the error should be logged with full traceback
    assert len(caplog.records) > 0
    error_message = caplog.records[0].message
    assert "Failed to load node from module invalid_node" in error_message
    assert "invalid_node.py, line 4" in error_message



from io import StringIO
import sys
import threading
from typing import Any, Callable, Optional

from werkzeug import local

# Save all of the objects for use later.
orig___stdout__ = sys.__stdout__
orig___stderr__ = sys.__stderr__
orig_stdout = sys.stdout
orig_stderr = sys.stderr
thread_proxies: dict[Any, Any] = {}


# Taken from https://stackoverflow.com/questions/14890997/redirect-stdout-to-a-file-only-for-a-specific-thread
# and updated for latest python version.
def redirect_log() -> StringIO:
    """
    Enables the redirect for the current thread's output to a single cStringIO
    object and returns the object.

    :return: The StringIO object.
    :rtype: ``cStringIO.StringIO``
    """
    # Get the current thread's identity.
    ident = threading.currentThread().ident

    # Enable the redirect and return the cStringIO object.
    thread_proxies[ident] = StringIO()
    return thread_proxies[ident]


def stop_redirect() -> Optional[str]:
    """
    Enables the redirect for the current thread's output to a single cStringIO
    object and returns the object.

    :return: The final string value.
    :rtype: ``str``
    """
    # Get the current thread's identity.
    ident = threading.currentThread().ident

    # Only act on proxied threads.
    if ident not in thread_proxies:
        return None

    # Read the value, close/remove the buffer, and return the value.
    retval = thread_proxies[ident].getvalue()
    thread_proxies[ident].close()
    del thread_proxies[ident]
    return retval


def _get_stream(original: Any) -> Callable[[], Any]:
    """
    Returns the inner function for use in the LocalProxy object.

    :param original: The stream to be returned if thread is not proxied.
    :type original: ``file``
    :return: The inner function for use in the LocalProxy object.
    :rtype: ``function``
    """

    def proxy() -> Any:
        """
        Returns the original stream if the current thread is not proxied,
        otherwise we return the proxied item.

        :return: The stream object for the current thread.
        :rtype: ``file``
        """
        # Get the current thread's identity.
        ident = threading.currentThread().ident

        # Return the proxy, otherwise return the original.
        return thread_proxies.get(ident, original)

    # Return the inner function.
    return proxy


def enable_log_proxy() -> None:
    """
    Overwrites __stdout__, __stderr__, stdout, and stderr with the proxied
    objects.
    """
    sys.__stdout__ = local.LocalProxy(_get_stream(sys.__stdout__))  # type: ignore
    sys.__stderr__ = local.LocalProxy(_get_stream(sys.__stderr__))  # type: ignore
    sys.stdout = local.LocalProxy(_get_stream(sys.stdout))  # type: ignore
    sys.stderr = local.LocalProxy(_get_stream(sys.stderr))  # type: ignore


def disable_log_proxy() -> None:
    """
    Overwrites __stdout__, __stderr__, stdout, and stderr with the original
    objects.
    """
    sys.__stdout__ = orig___stdout__
    sys.__stderr__ = orig___stderr__
    sys.stdout = orig_stdout
    sys.stderr = orig_stderr



import pytest

from vellum import (
    ChatMessage,
    FunctionCall,
    SearchResult,
    SearchResultDocument,
    StringVellumValue,
    VellumAudio,
    VellumError,
    VellumImage,
)
from workflow_server.utils.utils import (
    convert_json_inputs_to_vellum,
    to_python_safe_snake_case,
    to_valid_python_identifier,
)


@pytest.mark.parametrize(
    ["input", "expected"],
    [
        ({"type": "STRING", "name": "test", "value": "<example-string-value>"}, {"test": "<example-string-value>"}),
        ({"type": "NUMBER", "name": "test2", "value": 5}, {"test2": 5}),
        (
            {"type": "JSON", "name": "test3", "value": {"example-key": "example-value"}},
            {"test3": {"example-key": "example-value"}},
        ),
        (
            {
                "type": "CHAT_HISTORY",
                "name": "chat_history",
                "value": [{"role": "USER", "text": "<example-user-text>"}],
            },
            {"chat_history": [ChatMessage(text="<example-user-text>", role="USER")]},
        ),
        (
            {
                "type": "FUNCTION_CALL",
                "name": "function_call",
                "value": {"name": "example-function-name", "arguments": {"foo": "bar"}},
            },
            {"function_call": FunctionCall(name="example-function-name", arguments={"foo": "bar"})},
        ),
        (
            {"type": "IMAGE", "name": "image", "value": {"src": "https://example.com/image.png"}},
            {"image": VellumImage(src="https://example.com/image.png")},
        ),
        (
            {"type": "AUDIO", "name": "audio", "value": {"src": "https://example.com/audio.mp3"}},
            {"audio": VellumAudio(src="https://example.com/audio.mp3")},
        ),
        (
            {
                "type": "SEARCH_RESULTS",
                "name": "search_results",
                "value": [
                    {
                        "text": "example-search-result",
                        "score": 0.99,
                        "keywords": ["foo", "bar"],
                        "document": {
                            "id": "example-document-id",
                            "label": "example-document-label",
                            "external_id": "example-external-id",
                            "metadata": {"foo": "bar"},
                        },
                    }
                ],
            },
            {
                "search_results": [
                    SearchResult(
                        text="example-search-result",
                        score=0.99,
                        keywords=["foo", "bar"],
                        document=SearchResultDocument(
                            id="example-document-id",
                            label="example-document-label",
                            external_id="example-external-id",
                            metadata={"foo": "bar"},
                        ),
                    )
                ]
            },
        ),
        (
            {
                "type": "ERROR",
                "name": "error",
                "value": {"message": "example-error-message", "code": "USER_DEFINED_ERROR"},
            },
            {"error": VellumError(message="example-error-message", code="USER_DEFINED_ERROR")},
        ),
        (
            {"type": "ARRAY", "name": "array", "value": [{"type": "STRING", "value": "<example-string-value>"}]},
            {"array": [StringVellumValue(value="<example-string-value>")]},
        ),
        (
            {"type": "NUMBER", "name": "123", "value": 123},
            {"input_123": 123},
        ),
        (
            {"type": "STRING", "name": "_a", "value": "example-string-value"},
            {"input_a": "example-string-value"},
        ),
        (
            {"type": "NUMBER", "name": "_123", "value": 123},
            {"input_123": 123},
        ),
    ],
    ids=[
        "string",
        "number",
        "json",
        "chat_history",
        "function_call",
        "image",
        "audio",
        "search_results",
        "error",
        "array",
        "number_prefixed_input",
        "underscore_prefixed_alpha_input",
        "underscore_prefixed_number_input",
    ],
)
def test_convert_json_inputs_to_vellum__happy_path(input, expected):
    actual = convert_json_inputs_to_vellum([input])

    assert expected == actual


def test_input_variables_with_uppercase_gets_sanitized():
    inputs = [
        {"type": "STRING", "name": "Foo", "value": "<example-string-value>"},
        {"type": "STRING", "name": "Foo-Var", "value": "<another-example-string-value>"},
    ]

    expected = {
        "Foo": "<example-string-value>",
        "foo_var": "<another-example-string-value>",
    }

    actual = convert_json_inputs_to_vellum(inputs)

    assert expected == actual


@pytest.mark.parametrize(
    ["input_string", "safety_prefix", "expected"],
    [
        ("Foo", "input", "Foo"),
        ("test", "input", "test"),
        ("myVariable", "input", "myVariable"),
        ("validName123", "input", "validName123"),
        ("Foo-Var", "input", "foo_var"),
        ("My Variable", "input", "my_variable"),
        ("test-case", "input", "test_case"),
        ("CamelCase", "input", "CamelCase"),
        ("123", "input", "input_123"),
        ("_a", "input", "input_a"),
        ("_123", "input", "input_123"),
        ("test@#$", "input", "test"),
        ("@#$test", "input", "test"),
        ("123", "_", "_123"),
        ("123", "var", "var_123"),
        ("123", "var_", "var_123"),
    ],
)
def test_to_valid_python_identifier(input_string, safety_prefix, expected):
    actual = to_valid_python_identifier(input_string, safety_prefix)
    assert expected == actual


@pytest.mark.parametrize(
    ["input_string", "safety_prefix", "expected"],
    [
        ("Foo", "input", "foo"),
        ("Foo-Var", "input", "foo_var"),
        ("CamelCase", "input", "camel_case"),
        ("test", "input", "test"),
        ("My Variable", "input", "my_variable"),
        ("123", "input", "input_123"),
        ("_a", "input", "input_a"),
        ("_123", "input", "input_123"),
        ("123", "_", "_123"),
        ("123", "var", "var_123"),
        ("123", "var_", "var_123"),
    ],
)
def test_to_python_safe_snake_case(input_string, safety_prefix, expected):
    actual = to_python_safe_snake_case(input_string, safety_prefix)
    assert expected == actual



from uuid import uuid4

from workflow_server.server import create_app


def test_input_conversion_with_display_mapping():
    """
    Test that validates input conversion behavior for future WorkflowDisplay refactor.

    This test demonstrates the expected behavior when inputs are converted from UI names
    to SDK attribute names using WorkflowDisplay's inputs_display mapping, rather than
    the current snake_case conversion approach.

    The test passes on main branch by using the current conversion logic, but establishes
    the expected input/output patterns for the future refactor.
    """
    span_id = uuid4()
    request_body = {
        "execution_id": str(span_id),
        "inputs": [
            {"name": "User Message", "type": "STRING", "value": "Hello world"},
            {"name": "123-config", "type": "STRING", "value": "config-value"},
            {"name": "API-Key", "type": "STRING", "value": "test-key"},
        ],
        "environment_api_key": "test",
        "module": "workflow",
        "timeout": 360,
        "files": {
            "__init__.py": "",
            "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        result = "success"
""",
            "inputs.py": """
from vellum.workflows.inputs import BaseInputs

class Inputs(BaseInputs):
    user_message: str
    input_123_config: str
    api_key: str
""",
        },
    }

    flask_app = create_app()
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code

    assert status_code == 200, f"Request failed with status {status_code}: {response.data}"

    response_lines = [line for line in response.data.decode().split("\n") if line and line not in ["WAITING", "END"]]
    assert len(response_lines) > 0, "No response events received"

    fulfilled_events = [line for line in response_lines if "workflow.execution.fulfilled" in line]
    assert len(fulfilled_events) > 0, "No workflow.execution.fulfilled event received"



from datetime import datetime
import json
import logging
import os
from uuid import uuid4
from typing import Optional

from vellum.workflows.exceptions import WorkflowInitializationException
from workflow_server.api.workflow_view import get_workflow_request_context
from workflow_server.core.events import (
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
    VembdaExecutionInitiatedBody,
    VembdaExecutionInitiatedEvent,
)
from workflow_server.core.executor import stream_workflow
from workflow_server.core.workflow_executor_context import WorkflowExecutorContext
from workflow_server.utils.utils import get_version

logger = logging.getLogger(__name__)

_EVENT_LINE = "--event--"


def run_code_exec_stream() -> None:
    context: Optional[WorkflowExecutorContext] = None

    try:
        input_raw = ""
        while "--vellum-input-stop--" not in input_raw:
            # os/read they come in chunks of idk length, not as lines
            input_raw += os.read(0, 100_000_000).decode("utf-8")

        split_input = input_raw.split("\n--vellum-input-stop--\n")
        input_json = split_input[0]

        input_data = json.loads(input_json)
        context = get_workflow_request_context(input_data)

        print("--vellum-output-start--")  # noqa: T201

        initiated_event = VembdaExecutionInitiatedEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=context.trace_id,
            span_id=context.execution_id,
            body=VembdaExecutionInitiatedBody.model_validate(get_version()),
            parent=None,
        ).model_dump_json()

        print(f"{_EVENT_LINE}{initiated_event}")  # noqa: T201

        try:
            for line in stream_workflow(context, disable_redirect=True):
                print(f"{_EVENT_LINE}{json.dumps(line)}")  # noqa: T201
        except WorkflowInitializationException as e:
            fulfilled_event = VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=context.trace_id,
                span_id=context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr=str(e),
                    container_overhead_latency=context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")

            print(f"{_EVENT_LINE}{json.dumps(fulfilled_event)}")  # noqa: T201
    except Exception as e:
        logger.exception(e)

        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=context.trace_id if context else uuid4(),
            span_id=context.execution_id if context else uuid4(),
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                stderr="Internal Server Error",
            ),
            parent=None,
        )
        event = json.dumps(vembda_fulfilled_event.model_dump(mode="json"))
        print(f"{_EVENT_LINE}{event}")  # noqa: T201



from typing import Tuple

from flask import Blueprint, Response, jsonify

from workflow_server.utils.oom_killer import get_is_oom_killed

bp = Blueprint("healthz", __name__)


@bp.route("", methods=["GET"])
@bp.route("/", methods=["GET"])
def healthz() -> Tuple[Response, int]:
    if get_is_oom_killed():
        return jsonify(success=False), 500

    resp = jsonify(success=True)
    return resp, 200



from datetime import datetime
import importlib
from io import StringIO
import json
import logging
from multiprocessing import Process, Queue
import os
import random
import string
import sys
from threading import Event as ThreadingEvent
import time
from traceback import format_exc
from uuid import uuid4
from typing import Any, Callable, Generator, Iterator, Optional, Tuple, Type

from pebble import concurrent
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder

from vellum import Vellum, VellumEnvironment
from vellum.workflows import BaseWorkflow
from vellum.workflows.errors import WorkflowError, WorkflowErrorCode
from vellum.workflows.events.types import BaseEvent
from vellum.workflows.events.workflow import WorkflowEventDisplayContext
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode
from vellum.workflows.nodes.mocks import MockNodeExecution
from vellum.workflows.state.base import BaseState, StateMeta
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.state.store import EmptyStore
from vellum.workflows.workflows.event_filters import all_workflow_event_filter
from workflow_server.config import IS_VPC, VELLUM_API_URL_HOST, VELLUM_API_URL_PORT
from workflow_server.core.cancel_workflow import CancelWorkflowWatcherThread
from workflow_server.core.events import (
    STREAM_FINISHED_EVENT,
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
)
from workflow_server.core.workflow_executor_context import (
    DEFAULT_TIMEOUT_SECONDS,
    BaseExecutorContext,
    NodeExecutorContext,
    WorkflowExecutorContext,
)
from workflow_server.utils.log_proxy import redirect_log

logger = logging.getLogger(__name__)


@concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
# type ignore since pebble annotation changes return type
def stream_node_pebble_timeout(
    executor_context: NodeExecutorContext,
    queue: Queue,
) -> None:
    _stream_node_wrapper(
        executor_context=executor_context,
        queue=queue,
    )


def _stream_node_wrapper(executor_context: NodeExecutorContext, queue: Queue) -> None:
    try:
        for event in stream_node(executor_context=executor_context):
            queue.put(event)
    except Exception as e:
        logger.exception(e)
        queue.put(
            VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr="Internal Server Error",
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")
        )


def _stream_workflow_wrapper(executor_context: WorkflowExecutorContext, queue: Queue) -> None:
    try:
        for event in stream_workflow(executor_context=executor_context):
            queue.put(json.dumps(event))

    except WorkflowInitializationException as e:
        queue.put(
            VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr=str(e),
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")
        )
    except Exception as e:
        logger.exception(e)
        queue.put(
            VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr="Internal Server Error",
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")
        )
    queue.put(STREAM_FINISHED_EVENT)

    exit(0)


def stream_workflow_process_timeout(
    executor_context: WorkflowExecutorContext,
    queue: Queue,
) -> Process:
    workflow_process = Process(
        target=_stream_workflow_wrapper,
        args=(
            executor_context,
            queue,
        ),
    )
    workflow_process.start()

    if workflow_process.exitcode is not None:
        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                timed_out=True,
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        queue.put(vembda_fulfilled_event.model_dump(mode="json"))

    return workflow_process


def stream_workflow(
    executor_context: WorkflowExecutorContext,
    disable_redirect: bool = True,
) -> Iterator[dict]:
    workflow, namespace = _gather_workflow(executor_context)
    workflow_inputs = _get_workflow_inputs(executor_context)
    display_context = _gather_display_context(workflow, namespace)
    workflow_state = (
        workflow.deserialize_state(
            executor_context.state,
            workflow_inputs=workflow_inputs or BaseInputs(),
        )
        if executor_context.state
        else None
    )
    run_from_node = _get_run_from_node(executor_context, workflow)
    node_output_mocks = MockNodeExecution.validate_all(
        executor_context.node_output_mocks,
        workflow.__class__,
    )

    def call_workflow() -> Generator[dict[str, Any], Any, None]:
        cancel_watcher_kill_switch = ThreadingEvent()
        cancel_signal = ThreadingEvent()
        cancel_watcher = CancelWorkflowWatcherThread(
            kill_switch=cancel_watcher_kill_switch,
            execution_id=executor_context.execution_id,
            timeout_seconds=executor_context.timeout,
            vembda_public_url=executor_context.vembda_public_url,
            cancel_signal=cancel_signal,
        )

        try:
            if executor_context.vembda_public_url:
                cancel_watcher.start()

            stream = workflow.stream(
                inputs=workflow_inputs,
                state=workflow_state,
                node_output_mocks=node_output_mocks,
                event_filter=all_workflow_event_filter,
                cancel_signal=cancel_signal,
                entrypoint_nodes=[run_from_node] if run_from_node else None,
            )

            first = True
            for event in stream:
                if first:
                    executor_context.stream_start_time = time.time_ns()
                    first = False
                    if event.name == "workflow.execution.initiated":
                        event.body.display_context = display_context

                if event.name.endswith("rejected") and event.body.error.code.name == "INTERNAL_ERROR":  # type: ignore
                    logger.exception(event.body.error.message)  # type: ignore
                    event.body.error = WorkflowError(  # type: ignore
                        code=WorkflowErrorCode.INTERNAL_ERROR.value, message="Internal Error"  # type: ignore
                    )

                yield _dump_event(
                    event=event,
                    executor_context=executor_context,
                )
        finally:
            cancel_watcher_kill_switch.set()

    return _call_stream(
        executor_context=executor_context,
        stream_generator=call_workflow,
        disable_redirect=disable_redirect,
    )


def stream_node(
    executor_context: NodeExecutorContext,
    disable_redirect: bool = True,
) -> Iterator[dict]:
    namespace = _get_file_namespace(executor_context)

    def call_node() -> Generator[dict[str, Any], Any, None]:
        sys.meta_path.append(VirtualFileFinder(executor_context.files, namespace))
        workflow_context = _create_workflow_context(executor_context)
        node_module = importlib.import_module(f"{namespace}.{executor_context.node_module}")

        Node = getattr(node_module, executor_context.node_name)

        workflow_inputs = _get_workflow_inputs(executor_context)
        workflow_state = _get_workflow_state(executor_context, workflow_inputs=workflow_inputs)

        node = Node(
            state=workflow_state,
            context=workflow_context,
        )

        executor_context.stream_start_time = time.time_ns()
        node_outputs = node.run()

        if isinstance(node_outputs, (Iterator)):
            for node_output in node_outputs:
                yield json.loads(json.dumps(node_output, default=vars))
        else:
            yield json.loads(json.dumps(node_outputs, default=vars))

    return _call_stream(
        executor_context=executor_context,
        stream_generator=call_node,
        disable_redirect=disable_redirect,
    )


def _call_stream(
    executor_context: BaseExecutorContext,
    stream_generator: Callable[[], Generator[dict[str, Any], Any, None]],
    disable_redirect: bool = True,
) -> Iterator[dict]:
    log_redirect: Optional[StringIO] = None

    if not disable_redirect:
        log_redirect = redirect_log()

    try:
        yield from stream_generator()

        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=0,
                log=log_redirect.getvalue() if log_redirect else "",
                stderr="",
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        yield vembda_fulfilled_event.model_dump(mode="json")

    except Exception:
        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                log=log_redirect.getvalue() if log_redirect else "",
                stderr=format_exc(),
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        yield vembda_fulfilled_event.model_dump(mode="json")


def _create_workflow(executor_context: WorkflowExecutorContext, namespace: str) -> BaseWorkflow:
    workflow_context = _create_workflow_context(executor_context)
    Workflow = BaseWorkflow.load_from_module(namespace)
    VembdaExecutionFulfilledEvent.model_rebuild(
        # Not sure why this is needed, but it is required for the VembdaExecutionFulfilledEvent to be
        # properly rebuilt with the recursive types.
        _types_namespace={
            "BaseWorkflow": BaseWorkflow,
            "BaseNode": BaseNode,
        },
    )

    return Workflow(context=workflow_context, store=EmptyStore())


def _create_workflow_context(executor_context: BaseExecutorContext) -> WorkflowContext:
    if IS_VPC:
        environment = VellumEnvironment(
            default=os.getenv("VELLUM_DEFAULT_API_URL", VellumEnvironment.PRODUCTION.default),
            documents=os.getenv("VELLUM_DOCUMENTS_API_URL", VellumEnvironment.PRODUCTION.documents),
            predict=os.getenv("VELLUM_PREDICT_API_URL", VellumEnvironment.PRODUCTION.predict),
        )
    elif os.getenv("USE_LOCAL_VELLUM_API") == "true":
        VELLUM_API_URL = f"http://{VELLUM_API_URL_HOST}:{VELLUM_API_URL_PORT}"
        environment = VellumEnvironment(
            default=VELLUM_API_URL,
            documents=VELLUM_API_URL,
            predict=VELLUM_API_URL,
        )
    else:
        environment = VellumEnvironment.PRODUCTION

    if executor_context.environment_variables:
        os.environ.update(executor_context.environment_variables)

    return WorkflowContext(
        vellum_client=Vellum(
            api_key=executor_context.environment_api_key,
            environment=environment,
        ),
        execution_context=executor_context.execution_context,
        generated_files=executor_context.files,
    )


def _get_file_namespace(executor_context: BaseExecutorContext) -> str:
    return str(executor_context.execution_id) or "".join(
        random.choice(string.ascii_letters + string.digits) for i in range(14)
    )


def _dump_event(event: BaseEvent, executor_context: BaseExecutorContext) -> dict:
    module_base = executor_context.module.split(".")
    dump = event.model_dump(mode="json")
    if dump["name"] in {
        "workflow.execution.initiated",
        "workflow.execution.fulfilled",
        "workflow.execution.rejected",
        "workflow.execution.streaming",
        "workflow.execution.paused",
        "workflow.execution.resumed",
    }:
        dump["body"]["workflow_definition"]["module"] = module_base + dump["body"]["workflow_definition"]["module"][1:]
    elif dump["name"] in {
        "node.execution.initiated",
        "node.execution.fulfilled",
        "node.execution.rejected",
        "node.execution.streaming",
        "node.execution.paused",
        "node.execution.resumed",
    }:
        dump["body"]["node_definition"]["module"] = module_base + dump["body"]["node_definition"]["module"][1:]

    return dump


def _get_workflow_inputs(executor_context: BaseExecutorContext) -> Optional[BaseInputs]:
    if not executor_context.inputs:
        return None

    if not executor_context.files.get("inputs.py"):
        return None

    namespace = _get_file_namespace(executor_context)
    inputs_module_path = f"{namespace}.inputs"
    try:
        inputs_module = importlib.import_module(inputs_module_path)
    except Exception as e:
        raise WorkflowInitializationException(f"Failed to initialize workflow inputs: {e}") from e

    if not hasattr(inputs_module, "Inputs"):
        raise WorkflowInitializationException(
            f"Inputs module {inputs_module_path} does not have a required Inputs class"
        )

    if not issubclass(inputs_module.Inputs, BaseInputs):
        raise WorkflowInitializationException(
            f"""The class {inputs_module_path}.Inputs was expected to be a subclass of BaseInputs, \
but found {inputs_module.Inputs.__class__.__name__}"""
        )

    return inputs_module.Inputs(**executor_context.inputs)


def _get_workflow_state(
    executor_context: BaseExecutorContext, workflow_inputs: Optional[BaseInputs]
) -> Optional[BaseState]:
    namespace = _get_file_namespace(executor_context)
    State = importlib.import_module(f"{namespace}.state").State if executor_context.files.get("state.py") else None

    if not State:
        return None

    if not issubclass(State, BaseState):
        return None

    if executor_context.state:
        return State(**executor_context.state)

    if workflow_inputs:
        return State(
            meta=StateMeta(workflow_inputs=workflow_inputs),
        )

    return State()


def _get_run_from_node(executor_context: WorkflowExecutorContext, workflow: BaseWorkflow) -> Optional[Type[BaseNode]]:
    if not executor_context.node_id:
        return None

    for node in workflow.get_nodes():
        if node.__id__ == executor_context.node_id:
            return node

    return None


def _gather_workflow(context: WorkflowExecutorContext) -> Tuple[BaseWorkflow, str]:
    try:
        namespace = _get_file_namespace(context)
        sys.meta_path.append(VirtualFileFinder(context.files, namespace))
        workflow = _create_workflow(
            executor_context=context,
            namespace=namespace,
        )
        return workflow, namespace
    except Exception as e:
        logger.exception("Failed to initialize Workflow")
        raise WorkflowInitializationException(f"Failed to initialize workflow: {e}") from e


def _gather_display_context(workflow: BaseWorkflow, namespace: str) -> Optional["WorkflowEventDisplayContext"]:
    try:
        return BaseWorkflowDisplay.gather_event_display_context(namespace, workflow.__class__)
    except Exception:
        logger.exception("Unable to Parse Workflow Display Context")
        return None
